#include <iostream>

template <class D> class container {
    public:
        container *proximo;
        D elemento;
};


template <class T, class Container=container<T>>
class Fila {
    public:
        void enfilera(T elem) {
            N++;
            auto cont = new container<T>();
            cont->elemento = elem;

            if (fim != nullptr)
                fim->proximo = cont;
                
            else 
                inicio = cont;
            
            fim = cont;
        }


        T desenfilera() {
            if (inicio != nullptr) {
                N--;
                T elemento = inicio->elemento;
                Container *i = inicio;
                inicio = inicio->proximo;
                if (inicio == nullptr)
                    fim = nullptr;
                delete i;
                return elemento;
            }
            else
                return -1;
        }

        T frente() {
            if (inicio != nullptr)
                return inicio->elemento;
            else
                return -1;
        }


        int tamanho() {
            return N;
        }

        
    private:
        int N = 0;
        Container *inicio = nullptr;
        Container *fim = nullptr;

};

template <class T>
class Pilha { 

    public:
        void empilha(T elemento) {
            N++;
            elementos[N -1] = elemento;

        }
        T desempilha() {
            if (N > 0) {
                N--;
                return elementos[N];
            }
            else
                return -1;
        }

        T topo() {
            if (N > 0)
                return elementos[N -1];
            else   
                return -1;
        }

        int tamanho() {
            return N;
        }
    private:
        static const int MAX = 900;
        int N = 0;
        T elementos[MAX];

};

/**
 * @brief 
 * Para inverter uma fila utilizando duas filas:
 *  A idéia é percorrer toda a fila e pegando apenas o último elemento, e repetindo até finalizar
 * 1º Passo: Enfilero os itens em uma fila auxiliar, salvando o último elemento
 * 2° Passo: Enfilero o último elemento numa segunda fila auxiliar (essa fila que vai armazenar a fila invertida)
 * 3° Passo: Volto para a fila original todos os ítens, exceto o último elemento
 * 4º Passo: Verifico se chegou ao fim, se não chegou repito os passos
 * 
 * @tparam T 
 * @param fila 
 */

template <class T> void inverterFila(Fila<T> *fila) {

    Fila<T> filaAux, filaAux2;
    T ultimo;
    bool fim = false;

    while (!fim) {
        while (fila->frente() != -1) {
            ultimo = fila->frente();
            filaAux.enfilera(fila->desenfilera());

        }
        filaAux2.enfilera(ultimo);

        while (filaAux.frente() != -1 && filaAux.frente() != ultimo) {
            fila->enfilera(filaAux.desenfilera());
        }
        filaAux.desenfilera();

        if (fila->frente() == -1)
            fim = true;
    }

    while (filaAux2.frente() != -1) {
        fila->enfilera(filaAux2.desenfilera());
    }
}

/**
 * @brief Inverter uma fila utilizando uma pilha.
 * 
 * Simplesmente quando empilhamos os elementos da fila, e quandos desempilhamos, enfileirando
 * na fila, automaticamente os elementos estarão invertidos.
 * 
 * @tparam T 
 * @param fila 
 */
template <class T> void inverterFilaComPilha(Fila<T> *fila) {

    Pilha<T> pilha;
    while (fila->frente() != -1) {
        pilha.empilha(fila->desenfilera());
    }
    while (pilha.topo() != -1) {
        fila->enfilera(pilha.desempilha());
    }
}



void preencherFila(Fila<char> &fila) {

    fila.enfilera('a');
    fila.enfilera('b');
    fila.enfilera('c');
    fila.enfilera('d');
    fila.enfilera('e');
    fila.enfilera('f');
    fila.enfilera('g');
    fila.enfilera('h');
    fila.enfilera('i');
    fila.enfilera('j');
    fila.enfilera('k');
    fila.enfilera('l');
    fila.enfilera('m');
    fila.enfilera('n');
    fila.enfilera('o');
    fila.enfilera('p');
    fila.enfilera('q');
    fila.enfilera('r');
    fila.enfilera('s');
    fila.enfilera('t');
    fila.enfilera('u');
    fila.enfilera('v');
    fila.enfilera('w');
    fila.enfilera('x');
    fila.enfilera('y');
    fila.enfilera('z');
}

void printarFila(Fila<char> &fila, std::string texto) {

    std::cout << "----" << texto << "------" << std::endl;
    while (fila.frente() != -1) {
        std::cout << "| " << fila.desenfilera() << " " ;
    }
    std::cout << "|" << std::endl;
}

void printarPilha(Pilha<char> &pilha, std::string texto) {

    std::cout << "----" << texto << "------" << std::endl;
    while (pilha.topo() != -1) {
        std::cout << "| " << pilha.desempilha() << " " ;
    }
    std::cout << "|" << std::endl;
}

int main(int argc, char* argv[]) {

    Fila<char> fila;
    preencherFila(fila);
    printarFila(fila, "Fila normal");

    

    preencherFila(fila);
    inverterFila(&fila);
    printarFila(fila, "Fila invertida com duas filas");

    preencherFila(fila);
    inverterFilaComPilha(&fila);
    printarFila(fila, "Fila invertida com Pilha");
    

    
    return 0;
}